


import javax.swing.*;

import net.sourceforge.jcomponentbreadboard.JComponentBreadboard;

import java.awt.*;
import java.util.Arrays;


/**
 ** Produces a basic x-y chart by using the JComponentBreadboard
 ** facility to place variously colored JPanels within a grid of
 ** pixel-like cells.
 **
 **/ 

@SuppressWarnings("serial")
public class JBChart extends JComponentBreadboard {

// because MAXGRIDSIZE is protected, need this derived class to get at it
   static class GRID_BAG_LAYOUT extends GridBagLayout {
      public static int MAX_GRID_SIZE() {return MAXGRIDSIZE;};
   }

// The properties below, which have reasonable defaults, would typically not
// be changed from these defaults, which will be acceptable most of the time.   

/**  
 ** Sets the total number of x-grid-cells in the charting region (mapped to
 ** xMin...xMax). Effectively, defines the x-resolution of the chart.
 ** The total number of charted x-cells plus peripheral cells must not
 ** exceed GridBagLayout.MAXGRIDSIZE (currently 512), or an invalid state
 ** exception will be thrown when you create new JBCharts.
 **
 ** Certain properties, like that defined by setXSymbolsize(), are
 ** defined in terms of grid cells and thus actual screen x-dimension of
 ** symbols and similar chart aspects will change whenever xCells is
 ** changed.  Therefore, to preserve the relative size of symbols, etc.
 ** be sure to set xCells in concert with these other (static) properties.
 ** 
 **/
 // The panel/pixel x:y ratio 4:3 works well on my screen (and hopefully
 // on other screens, too). 512 is a GridBagLayout imposed max, and I
 // need to alow a few panels for axis labels and such, hence the
 // default choices shown below:
   public static final int DEFAULT_XCELLS = 480;
   public static final int DEFAULT_YCELLS = 360;
   private int xCells = DEFAULT_XCELLS;
   public void setXCells(int nCells) {
      assert(nCells > 0);
      assert(nCells < GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      xCells = nCells;
   }
   public int getXCells() {
      return xCells;
   }
   private int yCells = DEFAULT_YCELLS;
/**
 ** Analogous to setXCells(), but for the y-dimension of the chart
 **
 **/   
   public void setYCells(int nCells) {
      assert(nCells > 0);
      assert(nCells < GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      yCells = nCells;
   }
   public int getYCells() {
      return yCells;
   }

// widths that define x-y axes, their ticks and labels. All widths are in
// GridBagLayout cells, which are used sort of like scalable pixels.
//    
// The number of cells taken up by a tick or a symbol is usually small,
// so if we had made them a certain % of the total # x grid cells, etc.
// roundoff would play too big of a role in determining their size. So we
// just let the programmer set them in pixel units and choose a reasonable set
// of defaults that works well with the default gridsize. 
//   
// below "width" means the shorter edge (edge mapped to the smaller
// number of cells) and "length" the longer edge (edge mapped to the
// larger number of cells)
   public static final int DEFAULT_XAXISWIDTH = 1;
   private int xAxisWidth = DEFAULT_XAXISWIDTH;
   public void setXAxisWidth(int nCells) {
      assert(nCells >= 0);
      assert(nCells < GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      xAxisWidth = nCells;
   }
   public int getXAxisWidth() {
      return xAxisWidth;
   }
   public static final int DEFAULT_YAXISWIDTH = 1;
   private int yAxisWidth = DEFAULT_YAXISWIDTH;
   public void setYAxisWidth(int nCells) {
      assert(nCells >= 0);
      assert(nCells < GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      yAxisWidth = nCells;
   }
   public int getYAxisWidth() {
      return yAxisWidth;
   }

/**
 ** Sets/get half the width of an x-axis tick, in grid cell units.   
 **
 ** Tick widths are odd (2*nHalfWidth + 1) to assure proper centering.
 ** 
 **/
   public static final int DEFAULT_XHALFTICKWIDTH = 0;
   private int xHalfTickWidth = DEFAULT_XHALFTICKWIDTH;
   public void setXHalfTickWidth(int nHalfWidth) {
      assert(nHalfWidth >= 0);
      assert(2*nHalfWidth+1 < GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      xHalfTickWidth = nHalfWidth;
   }
   public int getXHalfTickWidth() {
      return xHalfTickWidth;
   }

/**
 ** Sets/get half the width of an y-axis tick, in grid cell units.   
 **
 ** Tick widths are odd (2*nHalfWidth + 1) to assure proper centering.
 ** 
 **/
   public static final int DEFAULT_YHALFTICKWIDTH = 0;
   private int yHalfTickWidth = DEFAULT_YHALFTICKWIDTH;
   public void setYHalfTickWidth(int nHalfWidth) {
      assert(nHalfWidth >= 0);
      assert(2*nHalfWidth+1 < GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      yHalfTickWidth = nHalfWidth;
   }
   public int getYHalfTickWidth() {
      return yHalfTickWidth;
   }

/**
 ** Sets/get half a graph symbol's x-dimension, in grid cell units.   
 **
 ** Symbol sizes are odd (2*nHalfSize + 1) to assure proper centering.
 ** 
 **/
   public static final int DEFAULT_XHALFSYMBOLSIZE = 3;
   private int xHalfSymbolSize = DEFAULT_XHALFSYMBOLSIZE;
   public void setXHalfSymbolSize(int nHalfWidth) {
      assert(nHalfWidth >= 0);
      assert(2*nHalfWidth+1 < GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      xHalfSymbolSize = nHalfWidth;
   }
   public int getXHalfSymbolSize() {
      return xHalfSymbolSize;
   }

   /**
    ** Sets/get half a graph symbol's y-dimension, in grid cell units.   
    **
    ** Symbol sizes are odd (2*nHalfSize + 1) to assure proper centering.
    ** 
    **/
   public static final int DEFAULT_YHALFSYMBOLSIZE = 3;
   private int yHalfSymbolSize = DEFAULT_YHALFSYMBOLSIZE;
   public void setYHalfSymbolSize(int nHalfWidth) {
      assert(nHalfWidth >= 0);
      assert(2*nHalfWidth+1 < GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      yHalfSymbolSize = nHalfWidth;
   }
   public int getYHalfSymbolSize() {
      return yHalfSymbolSize;
   }

// since gridlines are continuations of ticks, makes sense to synch their widths   
   private int getXHalfGridlineWidth() {
      return xHalfTickWidth;
   }
   private int getYHalfGridlineWidth() {
      return yHalfTickWidth;
   }

// types of gridline styles available:   
   public static final int GRIDLINES_NEVER = 0;             // remove all gridlines
   public static final int GRIDLINES_ON_LABELED_TICKS = 1;  // only on labeled ticks
   public static final int GRIDLINES_ON_ALL_TICKS = 2;      // even if label is null 

   public static final int GRIDLINES_X_DEFAULT = GRIDLINES_NEVER;
   public static final int GRIDLINES_Y_DEFAULT = GRIDLINES_NEVER;
// checks that gridlineStyle is valid
   private void checkGridlineStyle(int gridlineStyle) {
     assert(GRIDLINES_NEVER == gridlineStyle ||
                         GRIDLINES_ON_LABELED_TICKS == gridlineStyle ||
                         GRIDLINES_ON_ALL_TICKS == gridlineStyle);
   }
   


   
/**
 ** set/get length of x-axis ticks in y-grid-cell units
 **
 **/ 
   public static final int DEFAULT_XTICKLENGTH = 5;
   private int xTickLength = DEFAULT_XTICKLENGTH;
   public void setXTickLength(int nWidth) {
      assert(nWidth >= 0);
      assert(nWidth < GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      xTickLength = nWidth;
   }
   public int getXTickLength() {
      return xTickLength;
   }

/**
 ** set/get length of y-axis ticks in x-grid-cell units
 **
 **/ 
   public static final int DEFAULT_YTICKLENGTH = 5;
   private int yTickLength = DEFAULT_YTICKLENGTH;
   public void setYTickLength(int nWidth) {
      assert(nWidth >= 0);
      assert(nWidth < GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      yTickLength = nWidth;
   }
   public int getYTickLength() {
      return yTickLength;
   }


   
// The height of x-axis labels, width of y-axis labels, and width of the
// chart key, etc. only need a single grid-cell, which is resized
// independently of the grid-cells that define the x-y points on the
// chart (this is possible, given how GridBagLayout works, only because
// these elements are all positioned along the chart's edges).

   private final static int X_TICK_LABEL_HEIGHT = 1;  
   private final static int X_AXIS_LABEL_HEIGHT = 1;  
   private final static int TITLE_LABEL_HEIGHT = 1;  
   private final static int Y_TICK_LABEL_WIDTH = 1;  
   private final static int Y_AXIS_LABEL_WIDTH = 1;  
   private final static int N_CHARTKEY_WIDTH = 1; 

// combined width of the x-axis, x-axis tick mark, and x axis tick label
// in grid cell units
   private int getXAxisEnsembleWidth() {
      return xAxisWidth + xTickLength + X_TICK_LABEL_HEIGHT;
   }
// combined width of the y-axis, y-axis tick mark, and y axis tick label
// in grid cell units 
   private int getYAxisEnsembleWidth() {
      return yAxisWidth + yTickLength + Y_TICK_LABEL_WIDTH;
   }

/**
 ** The total number of x-grid cells used to represent the chart,
 ** including cells devoted to peripheral aspects of the chart, like
 ** tickmarks, axes labels, and the chart key
 **
 ** Must not exceed GridBagLayout.MAXGRIDSIZE (currently 512), or an
 ** exception will be thrown when instantiating JBCharts.
 **
 **/ 
   public int getXTotalCells() {
      return getYAxisEnsembleWidth() + Y_AXIS_LABEL_WIDTH + xCells + N_CHARTKEY_WIDTH;
   }
/** analogous to getXTotalCells(), but for the Y dimension */
   public int getYTotalCells() {
      return X_AXIS_LABEL_HEIGHT + getXAxisEnsembleWidth() + yCells + TITLE_LABEL_HEIGHT;
   }

// axis, tick, and gridline colors are fixed
   public static final Color AXIS_COLOR = Color.GRAY;
   public static final Color TICK_COLOR = AXIS_COLOR;
   public static final Color GRIDLINE_COLOR = AXIS_COLOR;
   
// returned as the row or col index for any double values outside of the charted range 
   private final int OFF_THE_CHART = Integer.MIN_VALUE; 

// the breadboard "grid" into which JComponents representing the chart are placed   
   private Object[][] breadboard;

// The properties below don't have defaults that would work for most users, and
//   normally would be defined by the user differently for each chart.
   
   private double[][] xData; // first curve's xData in xData[0], second in xData[1], etc.
   private double[][] yData; // first curve's yData in yData[0], second in yData[1], etc.
   private boolean[] curveVisibility; // is each curve visible, or hidden
   private Color[] curveColors; // color of each curve
   private boolean[] curveIsConnected;  // connect the symbols with straight lines?
   private JLabel[] curveLabels; // label (in chart key) for each curve
   private int[] curveXSizes; // half symbol size, in pixels; size[i]=2*curveSizes[i]+1
   private int[] curveYSizes; // half symbol size, in pixels; size[i]=2*curveSizes[i]+1
   private double xMin; // smallest x visible in charting region
   private double xMax;  // largest x visible in charting region
   private JLabel[] xLabels; // actual x-axis labels; size of array defines x-tick spacing
   private int xGridlineStyle=GRIDLINES_X_DEFAULT; // defines when x-ticks get gridlines
   private double yMin; // smallest y visible in charting region 
   private double yMax; // largest y visible in charting region 
   private JLabel[] yLabels; // actual y-axis labels; size of array defines y-tick spacing
   private int yGridlineStyle=GRIDLINES_Y_DEFAULT;// defines when y-ticks get gridlines

   private JLabel chartTitle = null;
   private JLabel yAxisLabel = null;
   private JLabel xAxisLabel = null;
   
   public void setXData(double[][] xData) {
      this.xData = xData;
   }
   public double[][] getXData() {
      return xData;
   }
   public void setYData(double[][] yData) {
      this.yData = yData;
   }
   public double[][] getYData() {
      return yData;
   }
   public void setCurveColors(Color[] curveColors) {
      this.curveColors = curveColors;
   }
   public Color[] getCurveColors() {
      return curveColors;
   }
   public void setCurveVisibility(boolean[] curveVisibility) {
      this.curveVisibility = curveVisibility;
   }
   public boolean[] getCurveVisibility() {
      return curveVisibility;
   }
   private int nVisibleCurves() {
      int result = 0;
      for (int i=0; i < curveVisibility.length; i++)
         if (curveVisibility[i]) result++;
      return result;
   }
   public void setCurveIsConnected(boolean[] curveIsConnected) {
      this.curveIsConnected = curveIsConnected;
   }
   public boolean[] getCurveIsConnected() {
      return curveIsConnected;
   }
   public void setCurveLabels(JLabel[] curveLabels) {
      this.curveLabels = curveLabels;
   }
   public JLabel[] getCurveLabels() {
      return curveLabels;
   }
   public void setCurveXSizes(int[] curveSizes) {
      this.curveXSizes = curveSizes;
   }
   public int[] getCurveXSizes() {
      return curveXSizes;
   }
   public void setCurveYSizes(int[] curveSizes) {
      this.curveYSizes = curveSizes;
   }
   public int[] getCurveSizes() {
      return curveYSizes;
   }
   public void setXMin(double xMin) {
      this.xMin = xMin;
   }
   public double getXMin() {
      return xMin;
   }
   public void setXMax(double xMax) {
      this.xMax = xMax;
   }
   public double getXMax() {
      return xMax;
   }
   public void setXLabels(JLabel[] xLabels) {
      this.xLabels = xLabels;
   }
   public JLabel[] getXLabels() {
      return xLabels;
   }
   public void setXGridlineStyle(int xGridlineStyle) {
      checkGridlineStyle(xGridlineStyle);
      this.xGridlineStyle = xGridlineStyle;
   }
   public int getXGridlineStyle() {
      return xGridlineStyle;
   }
   public void setYMin(double yMin) {
      this.yMin = yMin;
   }
   public double getYMin() {
      return yMin;
   }
   public void setYMax(double yMax) {
      this.yMax = yMax;
   }
   public double getYMax() {
      return yMax;
   }
   public void setYLabels(JLabel[] yLabels) {
      this.yLabels = yLabels;
   }
   public JLabel[] getYLabels() {
      return yLabels;
   }
   public void setYGridlineStyle(int yGridlineStyle) {
      checkGridlineStyle(yGridlineStyle);
      this.yGridlineStyle = yGridlineStyle;
   }
   public int getYGridlineStyle() {
      return yGridlineStyle;
   }

   public void setChartTitle(JLabel chartTitle) {
      this.chartTitle = chartTitle;
   }
   public JLabel getChartTitle() {
      return chartTitle;
   }
   public void setYAxisLabel(JLabel yAxisLabel) {
      this.yAxisLabel = yAxisLabel;
   }
   public JLabel getYAxisLabel() {
      return yAxisLabel;
   }
   public void setXAxisLabel(JLabel xAxisLabel) {
      this.xAxisLabel = xAxisLabel;
   }
   public JLabel getXAxisLabel() {
      return xAxisLabel;
   }

// creates a JPanel representing a colored "cell" of the given size, either on the chart,
// or as a symbol on a chart key   
   private JPanel newJPanel(Color color, Dimension minSize) {
      JPanel result = (JPanel) xyFill(new JPanel());
      result.setBackground(color);
//      result.setMinimumSize(minSize); 
      result.setPreferredSize(minSize); 
      return result;
   }
// creates a new JPanel representing a single chart "breadboard-cell-based pixel"
   private JPanel newPoint(Color color) {
      Dimension minSize = new Dimension(1,1);  // allows down to 1 pixel scaling
      return newJPanel(color, minSize);
   }
// returns a JPanel suitable for use as a chart key curve symbol for a
// curve of the given color, and having the given curve label
   private JPanel newChartSymbol(Color color, JLabel curveLabel) {
// a square block about half the height of the label seems about right for the key
      int halfHeight = 1+curveLabel.getMinimumSize().height/2;
      Dimension minSize = new Dimension(halfHeight, halfHeight);
      JPanel result = newJPanel(color, minSize);
      return result;
   }
// returns the writeable row index closest to the given y-value, or
// OFF_THE_CHART if the y value is outside of the charted y-range
   private int row(double y) {
      int result = OFF_THE_CHART;
      if (yMin <= y && y <= yMax) {
         double f = (y - yMin)/(yMax - yMin);
         if (f >= 1.0) f = 1.0 - Math.ulp(1.0);  // assure we are always a tad < 1.0
// It's conventional for row indexes to start at the top and increase
// going down (the opposite of charted y's, which grow from bottom to top)
         result = TITLE_LABEL_HEIGHT + yCells - 1 - (int) (f * yCells);
      }
      return result;
   }

// returns the writeable column index closest to the given x-value, or
// OFF_THE_CHART if the x value is outside of the charted x-range
   private int col(double x) {
      int result = OFF_THE_CHART;
      if (xMin <= x && x <= xMax) {
         double f = (x - xMin)/(xMax - xMin);
         if (f >= 1.0) f = 1.0 - Math.ulp(1.0);  // assure we are always a tad < 1.0
         result = Y_AXIS_LABEL_WIDTH + getYAxisEnsembleWidth()  + (int) (f * xCells);
      }
      return result;
   }

// places the component in the specified rectangular region on the breadboard
   private void placeComponent(JComponent jc,
                               int startCol, int endCol, int startRow, int endRow) {
      
      for (int iRow = startRow; iRow <= endRow; iRow++) {
         for (int iCol = startCol; iCol <= endCol; iCol++) {
           breadboard[iRow][iCol] = jc;
         }
      }
   }

// fill rectangular region with individual JPanel "point" objects of the given color
   private void fillWithPoints(Color color,
                               int startCol, int endCol, int startRow, int endRow) {
      
      for (int iRow = startRow; iRow <= endRow; iRow++) {
         for (int iCol = startCol; iCol <= endCol; iCol++) {
           breadboard[iRow][iCol] = newPoint(color);
         }
      }
   }

// fills the line between the two endpoints with a continous single-pixel line
   private void fillLineWithPoints(Color color, int c1, int r1, int c2, int r2) {
      int r;
      int c;
      if (r1 == r2 && c1 == c2) return;
      int rMin = Math.max(Math.min(r1,r2), row(yMax));
      int rMax = Math.min(Math.max(r1,r2), row(yMin));
      int cMin = Math.max(Math.min(c1,c2), col(xMin));
      int cMax = Math.min(Math.max(c1,c2), col(xMax));
             
      if (rMax - rMin > cMax - cMin) {
         for (r = rMin; r < rMax; r++) {
           double cDbl = c1 + (c2-c1)*(r-r1+0.0)/(r2-r1);
           c = (int) cDbl;           
           if (cMin <= c && c <= cMax) { 
// simple dithering strategy prevents "staircase look" of lines
        	 if (Math.random() >= cDbl-c)
               breadboard[r][c] = newPoint(color);
             else
                breadboard[r][c+1] = newPoint(color);
           }
         }
      }
      else {
         for (c = cMin; c < cMax; c++) {
           double rDbl = r1 + (r2-r1)*(c-c1+0.0)/(c2-c1);
           r = (int) rDbl;
           if (rMin <= r && r <= rMax) { 
             if (Math.random() >= rDbl-r) 
               breadboard[r][c] = newPoint(color);
             else
                breadboard[r+1][c] = newPoint(color);
           }
         }
      }

   }
   
/**
 ** Creates a JBChart with the data, x-y charted range, etc. as specified.
 **
 ** There is no way to change a JBChart's data, colors, etc. after creation; you
 ** need to create a completely new chart to effect such changes.
 ** 
 **/ 
   private void buildChartBreadboard() {

      assert(getYTotalCells() <= GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      assert(getXTotalCells() <= GRID_BAG_LAYOUT.MAX_GRID_SIZE());
      assert(xData.length == yData.length);
      assert(xData.length == curveColors.length);
      assert(xData.length == curveVisibility.length);
      assert(xData.length == curveIsConnected.length);
      assert(xData.length == curveLabels.length);
      assert(xData.length == curveXSizes.length);
      assert(xData.length == curveYSizes.length);
      for (int i = 0; i < xData.length; i++) {
         assert(xData[i].length == yData[i].length);
      }
      assert(xMax > xMin);
      assert(xLabels.length >= 2);
      checkGridlineStyle(xGridlineStyle);       
      assert(yMax > yMin);
      assert(yLabels.length >= 2);
      checkGridlineStyle(yGridlineStyle);
      
      breadboard = new Object[getYTotalCells()][getXTotalCells()];

// Create the x-axis and its associated tick-marks and tick-labels

// place x-axis itself

// To prevent data from overwriting the axes, axes are drawn one cell
// before the cell associated with the minimum charted x, y values.  The
// differences with the conventional approach where axes are drawn right
// on the minimum charted values won't usually be noticable.

// note: lower corner where x,y axes meet is part of the x-axis
      fillWithPoints(AXIS_COLOR,
                     Y_AXIS_LABEL_WIDTH + getYAxisEnsembleWidth()-yAxisWidth,
                     Y_AXIS_LABEL_WIDTH + getYAxisEnsembleWidth()+xCells-1,
                     yCells+TITLE_LABEL_HEIGHT, yCells+TITLE_LABEL_HEIGHT+xAxisWidth-1);

      double xGap = (xMax - xMin)/(xLabels.length-1);

      for (int iTick = 0; iTick < xLabels.length; iTick++) {
         double x = xMin + iTick*xGap;
        // these help define a region of cells centered on the tick
         int startCol = col(x-xGap/2);
         int midCol = col(x);
         int endCol = col(x+xGap/2);
// x and y axis labels closest to the origin can get mapped to the same
// rectangular grid cells when they are properly centered, violating the
// basic "one cell, one component" JComponentBreadboard constraint. To
// avoid having to deal with this and similar headaches, "off the chart"
// centered labels (which, usually, are the first and last labels on
// each axis) are simply dropped.
//
// Clients can partly work around this limitation by extending the chart
// range so that the last label they want to see becomes the next to last
// label on the label array, etc.
         
         if (OFF_THE_CHART != startCol && OFF_THE_CHART != endCol) {
        // place the tick itself
            fillWithPoints(TICK_COLOR,
                       midCol-xHalfTickWidth, midCol+xHalfTickWidth,
                           yCells+TITLE_LABEL_HEIGHT+xAxisWidth,
                           yCells+TITLE_LABEL_HEIGHT+xAxisWidth+xTickLength-1);
        // put out the gridline (if appropriate)
            if (GRIDLINES_ON_ALL_TICKS == xGridlineStyle ||
                (GRIDLINES_ON_LABELED_TICKS == xGridlineStyle &&
                 null != xLabels[iTick])) {
               fillWithPoints(GRIDLINE_COLOR,
                       midCol-getXHalfGridlineWidth(), midCol+getXHalfGridlineWidth(),
                       TITLE_LABEL_HEIGHT, yCells+TITLE_LABEL_HEIGHT-1);       
            }
            // place the label in a region below, and centered on, the tick
            if (xLabels[iTick] != null) {
               xLabels[iTick].setAlignmentX(0.5f); // center in its region along x
               xLabels[iTick].setHorizontalAlignment(SwingConstants.CENTER);
               xLabels[iTick].setHorizontalTextPosition(SwingConstants.CENTER);
// use the smallest integer difference to avoid "off by half cell" alignment errors
               int delta = Math.min(Math.abs(startCol-midCol), Math.abs(endCol-midCol));
               placeComponent(xLabels[iTick],
                              midCol-delta, midCol+delta,
                              yCells+TITLE_LABEL_HEIGHT+getXAxisEnsembleWidth()-X_TICK_LABEL_HEIGHT,
                              yCells+TITLE_LABEL_HEIGHT+getXAxisEnsembleWidth()-1);
            }
         }
      }

// Create the y-axis and its associated tick-marks and tick-labels

// place y-axis itself
// note: lower corner where x,y axes meet is part of the x-axis
      fillWithPoints(AXIS_COLOR,
                     Y_AXIS_LABEL_WIDTH+getYAxisEnsembleWidth()-yAxisWidth,
                     Y_AXIS_LABEL_WIDTH+getYAxisEnsembleWidth()-1,
                     TITLE_LABEL_HEIGHT, yCells+TITLE_LABEL_HEIGHT-1);

      double yGap = (yMax - yMin)/(yLabels.length-1);

      for (int iTick = 0; iTick < yLabels.length; iTick++) {
         double y = yMin + iTick*yGap;
        // these help define a region of cells centered on the tick
         int startRow = row(y+yGap/2);  // note: row-order is opposite y-order
         int midRow = row(y);
         int endRow = row(y-yGap/2);
// see note above regarding dropped axis labels in analogous section of x-axis code.
         if (OFF_THE_CHART != startRow && OFF_THE_CHART != endRow) {
        // create the tick itself
            fillWithPoints(TICK_COLOR,
                           Y_AXIS_LABEL_WIDTH+Y_TICK_LABEL_WIDTH,
                           Y_AXIS_LABEL_WIDTH+Y_TICK_LABEL_WIDTH+yTickLength-1,
                           midRow-yHalfTickWidth, midRow+yHalfTickWidth);
        // put out the gridline (if appropriate)
            if (GRIDLINES_ON_ALL_TICKS == yGridlineStyle ||
                (GRIDLINES_ON_LABELED_TICKS == yGridlineStyle &&
                 null != yLabels[iTick])) {
               fillWithPoints(GRIDLINE_COLOR,
                              Y_AXIS_LABEL_WIDTH+getYAxisEnsembleWidth(),
                              Y_AXIS_LABEL_WIDTH+getYAxisEnsembleWidth()+xCells-1,
                       midRow-getYHalfGridlineWidth(), midRow+getYHalfGridlineWidth());
            }
            // place the label in a region below, and centered on, the tick
            if (yLabels[iTick] != null) {
               yLabels[iTick].setAlignmentY(0.5f); // center in its region along y
               yLabels[iTick].setVerticalAlignment(SwingConstants.CENTER);
               yLabels[iTick].setVerticalTextPosition(SwingConstants.CENTER);
// use the smallest integer difference to avoid "off by half cell" alignment errors
               int delta = Math.min(Math.abs(startRow-midRow), Math.abs(endRow-midRow));
               placeComponent(yLabels[iTick],
                              Y_AXIS_LABEL_WIDTH,
                              Y_AXIS_LABEL_WIDTH+Y_TICK_LABEL_WIDTH-1, 
                              midRow-delta, midRow+delta);
            }
         }
      }
      
// Create the actual graph curves

      for (int iCurve = 0; iCurve < xData.length; iCurve++) {
          int c=0,cPrev=0;
          int r=0,rPrev=0;
        if (curveVisibility[iCurve]) {  
           for (int iPoint = 0; iPoint < xData[iCurve].length; iPoint++) {
               c = col(xData[iCurve][iPoint]);
               r = row(yData[iCurve][iPoint]);
               if (c != OFF_THE_CHART && r != OFF_THE_CHART) {
                 fillWithPoints(curveColors[iCurve],
                                Math.max(c-curveXSizes[iCurve],
                                         Y_AXIS_LABEL_WIDTH+getYAxisEnsembleWidth()),
                                Math.min(c+curveXSizes[iCurve],
                                     Y_AXIS_LABEL_WIDTH+getYAxisEnsembleWidth()+xCells-1),
                                Math.max(r-curveYSizes[iCurve],TITLE_LABEL_HEIGHT),
                                Math.min(r+curveYSizes[iCurve],
                                         TITLE_LABEL_HEIGHT+yCells-1));
               }

               if (iPoint > 0 && curveIsConnected[iCurve]) {
                  fillLineWithPoints(curveColors[iCurve], cPrev, rPrev, c, r);
               }
               cPrev = c;
               rPrev = r;
           }
        }
      }
      
// Create and add the embedded chart key as an embedded JComponentBreadboard
// (using an embedded breadboard allows the key to be centered, and its
// rows spaced, independently of the rows representing the chart proper)
     
      Object[][] chartKeyBreadboard = new Object[nVisibleCurves()][3];
      int iKeyPosition = 0;
      for (int iCurve = 0; iCurve < curveLabels.length; iCurve++) {
        if (curveVisibility[iCurve]) {  
       // the bracketing panels put a little space around symbol in the middle
           JPanel p1 = newChartSymbol(curveLabels[iCurve].getBackground(), curveLabels[iCurve]);
           JPanel p2 = newChartSymbol(curveColors[iCurve], curveLabels[iCurve]);
           JPanel p3 = newChartSymbol(curveLabels[iCurve].getBackground(), curveLabels[iCurve]);
           chartKeyBreadboard[iKeyPosition][0] = yAlign(0.5,new JComponentBreadboard(new Object[][] 
             {{p1, p2, p3}}
            ));
            chartKeyBreadboard[iKeyPosition][1] = yAlign(0.5,curveLabels[iCurve]);
            iKeyPosition++;
        // else both symbol and label will be null, which deletes hidden curve keys
        }        
      }
      if (chartKeyBreadboard.length > 0 ) {   
        JComponent chartKey = new JComponentBreadboard(chartKeyBreadboard);
        chartKey.setAlignmentX(0.5f);
        chartKey.setAlignmentY(0.25f);
        placeComponent(chartKey,
                     xCells+getYAxisEnsembleWidth()+Y_AXIS_LABEL_WIDTH,
                     xCells+getYAxisEnsembleWidth()+Y_AXIS_LABEL_WIDTH+N_CHARTKEY_WIDTH-1,
         TITLE_LABEL_HEIGHT, TITLE_LABEL_HEIGHT+yCells-1);
      }
      
      placeComponent(chartTitle,
                     0, getXTotalCells()-1,0, 0);
      placeComponent(yAxisLabel,0,Y_AXIS_LABEL_WIDTH-1,
                     TITLE_LABEL_HEIGHT, TITLE_LABEL_HEIGHT+yCells-1);
      placeComponent(xAxisLabel,
                     getYAxisEnsembleWidth()+Y_AXIS_LABEL_WIDTH,
                     getYAxisEnsembleWidth()+Y_AXIS_LABEL_WIDTH+xCells-1,
                     getYTotalCells()-X_AXIS_LABEL_HEIGHT,getYTotalCells()-1);

   }

   
   private int currentChartHashCode = 0;
   private boolean chartNeedsUpdate = true;
   private int chartHashCode() {
	      final int N_INPUTS = 26;  //must equal number of "tracked" inputs
	      Object[] inputs = new Object[N_INPUTS];
	      int iInput = 0;
	      inputs[iInput++] = xCells;
	      inputs[iInput++] = yCells;
	      inputs[iInput++] = xAxisWidth;
	      inputs[iInput++] = yAxisWidth;
	      inputs[iInput++] = yHalfTickWidth;
	      inputs[iInput++] = xHalfSymbolSize;
	      inputs[iInput++] = yHalfSymbolSize;
	      inputs[iInput++] = xTickLength;
	      inputs[iInput++] = yTickLength;
	      inputs[iInput++] = xData;
	      inputs[iInput++] = yData;
	      inputs[iInput++] = curveColors;
	      inputs[iInput++] = curveLabels;
	      inputs[iInput++] = curveXSizes;
	      inputs[iInput++] = curveYSizes;
	      inputs[iInput++] = xMin;
	      inputs[iInput++] = xMax;
	      inputs[iInput++] = xLabels;
	      inputs[iInput++] = xGridlineStyle;
	      inputs[iInput++] = yMin;
	      inputs[iInput++] = yMax;
	      inputs[iInput++] = yLabels;
	      inputs[iInput++] = yGridlineStyle;
	      inputs[iInput++] = chartTitle;
	      inputs[iInput++] = yAxisLabel;
	      inputs[iInput++] = xAxisLabel;
	      int result = Arrays.deepHashCode(inputs);
	      return result;
	   }
/**
 ** Returns just the JComponentBreadboard array representing the chart. Useful
 ** when constructing "connected" JBCharts within another, parent, JComponentBreadboard
 ** 
 **/ 
   public Object[][] getChartBreadboard() {
      int newHashCode = chartHashCode();
      if (newHashCode != currentChartHashCode) {
         buildChartBreadboard();
         currentChartHashCode = newHashCode;
         chartNeedsUpdate = true;
      }
      return breadboard;
   }
   
/**
 ** By adding appropriate JPanels, labels, etc. constructs an x-y chart consistent with 
 ** current specifications, and then intantiate that chart in this JPanel.
 ** 
 **/   
   public void updateChart() {
      breadboard = getChartBreadboard(); //changes chartNeedsUpdate if breadboard changes
      if (chartNeedsUpdate) {
         setBreadboard(breadboard);
 //        validate();
         chartNeedsUpdate = false;
      }
      jbRefresh();
   }
   
   public JBChart() {
      super();
   }

   
   static private JFrame showInAFrame(JPanel p) {    
	     JFrame jf = new JFrame();
		  jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		  jf.getContentPane().add(p);
		  jf.setSize(new Dimension(600,600));
//		  jf.validate();
		  jf.pack();
		  jf.setVisible(true);
		  return jf;
	 }
   public static void test_01() {
	// next line is "run this in the Swing thread" Java API boilerplate
	SwingUtilities.invokeLater(new Runnable() { public void run() {

      JBChart myChart = new JBChart();
//      myChart.setXCells(2*myChart.getXCells()/3);
//      myChart.setYCells(2*myChart.getYCells()/3);
      myChart.setXData(new double[][]
                       {{0, 1, 2, 3, 4},
    		               {0,0.5,1,1.5,2,2.5,3,3.5,4.0}});

      myChart.setYData(new double[][]
                       {{1.0, 2.0, 4.0, 8.0, 16.0},
    		              {2.0,3,4,5,6,7,8.1,9,10}});

      myChart.setCurveVisibility(new boolean[] {true, true});
      myChart.setCurveColors(new Color[] {Color.RED, Color.GREEN});
      myChart.setCurveIsConnected(new boolean[] {true, false});
      myChart.setCurveLabels(new JLabel[] {new JLabel("Red"), new JLabel("Green")});
      myChart.setCurveXSizes(new int[] {2, 4});
      myChart.setCurveYSizes(new int[] {2, 4});
      myChart.setXMin(0.0);
      myChart.setXMax(5.0);
      myChart.setXLabels(new JLabel[]  {new JLabel("0"),new JLabel("1"), new JLabel("2"),
                          new JLabel("3"), new JLabel("4"), new JLabel("5")});
      myChart.setYMin(0.0);
      myChart.setYMax(20.0);
      myChart.setYLabels(new JLabel[] {new JLabel(" 0"),new JLabel(" 5"), new JLabel("10"),
                         new JLabel("15"), new JLabel("20")});
//      myChart.setXGridlineStyle(JBChart.GRIDLINES_ON_ALL_TICKS);
//      myChart.setYGridlineStyle(JBChart.GRIDLINES_ON_ALL_TICKS);
      JLabel title = new JLabel("<HTML><H1>Two Curve Test Chart</H1>");
//      title.setVerticalAlignment(JLabel.BOTTOM);
      title.setAlignmentX(JLabel.CENTER_ALIGNMENT);
      myChart.setChartTitle(title);
      JLabel xAxisLabel = new JLabel("X Axis");
      xAxisLabel.setAlignmentX( JLabel.CENTER_ALIGNMENT);
      xAxisLabel.setHorizontalTextPosition(JLabel.CENTER);
      myChart.setXAxisLabel(xAxisLabel);
      JLabel yAxisLabel = new JLabel("<HTML><PRE>Y\n\nA \nx \ni \ns ");
      yAxisLabel.setHorizontalAlignment( JLabel.LEFT);
      yAxisLabel.setHorizontalTextPosition( JLabel.LEFT);
      myChart.setYAxisLabel(yAxisLabel);
                                                                         
      myChart.updateChart();
      JFrame jf = showInAFrame(myChart);
      boolean color1 = false;
      boolean firstVisible = true;
      boolean secondVisible = true;
      while (JOptionPane.YES_OPTION == JOptionPane.showConfirmDialog(myChart,"Curve color toggle test in  test_01 ... Toggle Again?","",JOptionPane.YES_NO_OPTION)) {
//      title.setHorizontalAlignment(JLabel.LEFT);
    	if (color1) {
          myChart.setCurveColors(new Color[] {Color.RED, Color.GREEN});
          myChart.setXMax(myChart.getXMax()/2);
          myChart.setCurveVisibility(new boolean[] {firstVisible, secondVisible});
          firstVisible = !firstVisible;
    	}  
    	else {
          myChart.setCurveColors(new Color[] {Color.BLUE, Color.CYAN});
          myChart.setXMax(myChart.getXMax()*2);
          myChart.setCurveVisibility(new boolean[] {firstVisible, secondVisible});
          secondVisible = !secondVisible;
    	}
    	color1 = !color1;
        myChart.updateChart();
        jf.validate(); 
        jf.repaint();
      }
	}}); // end of "run this in the Swing thread" section.
   }

   // returns a hash code associated with the properties that define the chart
 // this test sets every "JPanel pixel" with the largest possible grid-size, just
// to see how much memory this requires
   public static void test_02() {
	// next line is "run this in the Swing thread" Java API boilerplate
	SwingUtilities.invokeLater(new Runnable() { public void run() {
      JBChart myChart = new JBChart();
      final int xN = DEFAULT_XCELLS;
      final int yN = DEFAULT_YCELLS;
      double[][] xData = new double[1][xN*yN];
      double[][] yData = new double[1][xN*yN];
 // most curves will fill significantly less than 5% of the available chart points
      final double FRACTION_FILLED = 0.05;
      for (int i = 0; i < xN; i++) {
        for (int j = 0; j < yN; j++) {
           if (Math.random() < FRACTION_FILLED) {
             xData[0][i*yN+j] = i;
             yData[0][i*yN+j] = j;
           }
           else {
             xData[0][i*yN+j] = -xN;   // off the chart, so no point seen 
             yData[0][i*yN+j] = -yN;        	   
           }
        }
      }
     myChart.setXData(xData);
     myChart.setYData(yData);
     
     myChart.setCurveVisibility(new boolean[]  {true});
     myChart.setCurveColors(new Color[]  {Color.RED});
     myChart.setCurveIsConnected(new boolean[] {false});
     myChart.setCurveLabels(new JLabel[]  {new JLabel("Red")});
     myChart.setCurveXSizes(new int[] {0});
     myChart.setCurveYSizes(new int[] {0});
     myChart.setXMin(0);
     myChart.setXMax(xN);
     myChart.setXLabels(new JLabel[]  {new JLabel("0"),new JLabel("100"), new JLabel("200"),
                          new JLabel("300"), new JLabel("400"), new JLabel("500")});
     myChart.setYMin(0);
     myChart.setYMax(yN);
     myChart.setYLabels(new JLabel[]  {new JLabel("0"),new JLabel("100"), new JLabel("200"),
                          new JLabel("300"), new JLabel("400"), new JLabel("500")});
     JLabel myTitle = (JLabel) JComponentBreadboard.xAlign(0.5,
       new JLabel("<HTML><H1>5% of Points Filled at Random<BR><CENTER>(test chart 2)</CENTER>"));
     myChart.setChartTitle(myTitle);
     JLabel myXAxisLabel = (JLabel) JComponentBreadboard.xAlign(0.5,
    		 new JLabel("X"));
     myChart.setXAxisLabel(myXAxisLabel);
     myChart.setYAxisLabel(new JLabel("   Y"));
     myChart.updateChart();
     showInAFrame(myChart);
     myTitle.setHorizontalAlignment(JLabel.CENTER);
     
	}}); // end of "run this in the Swing thread" section.
   }
   public static void main(String[] args) {
      test_01();
      test_02();
   }
}